c23456789 123456789 123456789 123456789 123456789 123456789 123456789 12
C
C     Fortran program maxtrace to
C
c     read a sequence of 3x3 matrices, and for each matrix A in
c     the sequence to compute a 3x3 rotation matrix R (R orthogonal
c     of determinant equal to 1) so that RA is of maximal trace over
c     rotation matrices without the use of the SVD (singular value
c     decomposition) if possible.
c
c     A dxd matrix M is of maximal trace over rotation matrices if
c     given any dxd rotation matrix U (U orthogonal of determinant
c     equal to 1), then the trace of UM does not exceed that of M.
c
c     It is well known that given any dxd matrix A for any d, a dxd
c     rotation matrix R can be computed with a method based on the
c     computation of the SVD of A so that RA is of maximal trace
c     over rotation matrices.
c
c     With A a 3x3 matrix in the sequence read by the program,
c     if possible this program computes R without the SVD.
c     If $A$ is a 3x3 symmetric matrix that is not of maximal trace
c     over rotation matrices (this is checked by the program), then
c     in order to obtain a 3x3 rotation matrix R without the SVD so
c     that RA is of maximal trace over rotation matrices, it suffices
c     to compute R = 2rr^T-I, where I is the 3x3 identity matrix,
c     and r is a unit (column) vector in R^3 that is an eigenvector
c     of A corresponding to the largest eigenvalue of A.
c     Computing the eigenvalues of A is accomplished in this program
c     by taking advantage of a 3x3 matrix B that is a linear
c     combination of $A$ and $I$ in the appropriate manner so that
c     the characteristic polynomial of $B$ is such that it allows the
c     application of a trigonometric identity in order to obtain its
c     roots in closed form and thus those of the characteristic
c     polynomial of A.
c     On the other hand, if the matrix A is not symmetric, part of the
c     approach just described can still be used by the program on A^TA
c     which is symmetric (A^T is the transpose of A), to produce the
c     usual orthogonal matrices necessary to carry out the SVD method.
c     Finally, if the matrix A is not symmetric, the option exists in
c     the program to apply a procedure that uses the so-called Cayley
c     transform in conjunction with Newton's method to find a 3x3
c     rotation matrix $U$ so that $UA$ is symmetric, possibly of
c     maximal trace over rotation matrices. If the resulting $UA$ is
c     not of maximal trace over rotation matrices, using the fact that
c     UA is symmetric, another 3x3 rotation matrix R can then be
c     computed (without the SVD) as described above so that RUM is of
c     maximal trace over rotation~matrices. Since Newton's method can
c     fail, whenever this happens, as a last resort the SVD method can
c     then be used as described above.
c     Note that all results just mentioned are justified in a paper
c     still in preparation of tentative title "Characterization and
c     Computation of Matrices of Maximal Trace over Rotations" by
c     James F. Lawrence and Javier Bernal.
c
c     Here integer variable svdonly is set to 1 if only the SVD
c     (singular value decomposition) is to be used, set to something
c     else otherwise. Integer variable itex is set to the maximum number
c     of iterations of Newton's method allowed (if it is being used,
c     i.e., svdonly is not set to 1).
c
c     For each matrix A in the sequence read by the program, the
c     program computes a 3x3 rotation matrix R, a 3x3 matrix S such
c     that S = RA and S is of maximal trace over rotation matrices,
c     and the trace trc of S.
c
c     Disclaimer:
c
c     Author: Javier Bernal
c             National Institute of Standards and Technology (NIST)
c
c     This software was developed at the National Institute of Standards
c     and Technology by employees of the Federal Government in the
c     course of their official duties. Pursuant to title 17 Section 105
c     of the United States Code this software is not subject to
c     copyright protection and is in the public domain. This software is
c     experimental. NIST assumes no responsibility whatsoever for its
c     use by other parties, and makes no guarantees, expressed or
c     implied, about its quality, reliability, or any other
c     characteristic. We would appreciate acknowledgement if the
c     software is used.
c
*MAIN
c
      program main
c     
      double precision A(3,3), R(3,3), S(3,3), trc, d2, d3
      integer svdonly, itex, iter, i, j, k
      integer imnum
      INTEGER ICOUNT1, ICOUNT2, ICOUNT3, ICOUNT4
      INTEGER ITAR
c
c     initialize
c
      svdonly = 0
      itex = 60
c
      ICOUNT1 = 0
      ICOUNT2 = 0
      ICOUNT3 = 0
      ICOUNT4 = 0
      ITAR = 0
c
c     read matrix A
c
      write(*,*)'*******************************************'
      imnum = 0
      open(11, file='randomtrix')
   50 continue
      imnum = imnum+1
      read(11,*,end=7000) A(1,1), A(1,2), A(1,3)
      read(11,*) A(2,1), A(2,2), A(2,3)
      read(11,*) A(3,1), A(3,2), A(3,3)
      WRITE(*,*)'MATRIX NUMBER =',IMNUM
c     write(*,*)'Matrix A:'
c     do 100 i = 1, 3
c        write(*,*)(A(i,j),j=1,3)
c 100 continue
c
c------------------------------------------------------------
c
      call maxtrace(A, R, svdonly, itex, iter,
     *              ICOUNT1, ICOUNT2, ICOUNT3, ICOUNT4)
      ITAR = ITAR + ITER
c
c-------------------------------------------------------------
c
c
c     compute RA which should be of maximal trace over rotations
c
      do 4800 i = 1, 3
         do 4700 j = 1, 3
            S(i,j) = 0.0d0
            do 4600 k = 1, 3
               S(i,j) = S(i,j) + R(i,k)*A(k,j)
 4600       continue
 4700    continue
 4800 continue
c     write(*,*)'4900 Matrix S:'
c     do 4900 i = 1, 3
c        write(*,*)(S(i,j),j=1,3)
c4900 continue
      trc = S(1,1)+S(2,2)+S(3,3)
      WRITE(*,*)'TRACE=',TRC
c
      write(*,*)'****************************'
      go to 50
 7000 continue
      imnum = imnum-1
      WRITE(*,*)'TOTAL # OF MATRICES =',IMNUM
      IF(SVDONLY.EQ.1) GO TO 8000
      WRITE(*,*)'# OF SYMMETRY SUCCESSES =',ICOUNT1
      WRITE(*,*)'# OF SYMMETRY FAILURES =',ICOUNT2
      WRITE(*,*)'# OF SYMMETRY SUCCESSES OF MAXIMAL TRACE =',ICOUNT3
      WRITE(*,*)'# OF SYMMETRY SUCCESSES NOT OF MAXIMAL TRACE =',ICOUNT4
      WRITE(*,*)'TOTAL NUMBER OF NEWTON METHOD ITERATIONS =',ITAR
      WRITE(*,*)'MAXIMUM NUMBER OF NEWTON METHOD ITERATIONS ALLOWED ',
     *          'PER MATRIX =',ITEX
      D2 = DBLE(ITAR)
      D3 = DBLE(IMNUM)
      D2 = D2/D3
      WRITE(*,*)'AVERAGE # OF ITERATIONS OF NEWTON METHOD PER ',
     *          'MATRIX =', D2
 8000 CONTINUE
      write(*,*)' '
c
c
      stop
      end
c
*MAXTRACE
c
c     Program maxtrace
c
c     Compute matrix of maximal trace in 3d without the SVD if possible
c
      subroutine maxtrace(A, R, svdonly, itex, iter,
     *                    ICOUNT1, ICOUNT2, ICOUNT3, ICOUNT4)
c
      double precision A(3,3), M(3,3), ATA(3,3)
      double precision S(3,3), R(3,3), F(3,3)
      double precision epn, epp, atrace, g(3), h(3)
      double precision Fr(3,3), Fs(3,3), Ft(3,3), JJ(3,3), JI(3,3)
      double precision d2, d3, delta2
      double precision rr, ss, tt, rr2, ss2, tt2, rs, rt, st
      double precision sig1, sig2, sig3, sig(3), w(3,3), u(3,3), r3(3)
      integer i, j, k, icode, iter, itex, iflag, svdonly
      INTEGER ICOUNT1, ICOUNT2, ICOUNT3, ICOUNT4
c
c     set tolerances
c
      epn = -0.00001d0
      epp =  0.00001d0
      iter = 0
c
c     test for symmetry
c
      if(A(1,2)-A(2,1).le.epn .or. A(1,2)-A(2,1).ge.epp) go to 2000
      if(A(1,3)-A(3,1).le.epn .or. A(1,3)-A(3,1).ge.epp) go to 2000
      if(A(2,3)-A(3,2).le.epn .or. A(2,3)-A(3,2).ge.epp) go to 2000
      iflag = 0
c
c     write(*,*)'Matrix A appears to be symmetric'
c
c     test for positive semidefiniteness
c     set up matrix S = tr(A)I-A
c
  120 continue
      do 300 i = 1, 3
         do 200 j = 1, 3
            S(i,j) = -A(i,j)
  200    continue
  300 continue
      atrace = A(1,1)+A(2,2)+A(3,3)
      do 500 i=1,3
         S(i,i) = S(i,i) + atrace
  500 continue
c     write(*,*)'Matrix S:'
c     do 600 i = 1, 3
c        write(*,*)(S(i,j),j=1,3)
c 600 continue
c
c     test for positive definiteness and semidefiniteness of S
c
      icode = 1
      if(S(1,1).le.epn .or. S(2,2).le.epn .or. S(3,3).le.epn)
     *   go to 1000
      d2 = S(1,1)*S(2,2)-S(1,2)*S(2,1)
      icode = 2
      if(d2.le.epn) go to 1000
      d3 = S(1,1)*S(2,2)*S(3,3) + S(2,1)*S(3,2)*S(1,3) +
     *     S(1,2)*S(2,3)*S(3,1) - S(1,3)*S(2,2)*S(3,1) -
     *     S(1,2)*S(2,1)*S(3,3) - S(2,3)*S(3,2)*S(1,1)
c     WRITE(*,*)'D3=',D3
      icode = 3
      if(d3.le.epn) go to 1000
      if(S(1,1).ge.epp .and. d2.ge.epp .and. d3.ge.epp) then
c        write(*,*)'Matris S appears to be positive definite'
c        write(*,*)'Thus matrix A is of maximal trace over rotations'
         go to 800
      endif
c
      d2 = S(1,1)*S(3,3)-S(1,3)*S(3,1)
      icode = 4
      if(d2.le.epn) go to 1000
      d2 = S(2,2)*S(3,3)-S(2,3)*S(3,2)
      icode = 5
      if(d2.le.epn) go to 1000
c     write(*,*)'Matrix S appears to be positive semidefinite'
c     write(*,*)'Thus matrix A is of maximal trace over rotations'
c
  800 continue
      if(iflag.ne.0) go to 900
c
c     optimal rotation is taken to be the identity matrix
c
      do 830 i = 1, 3
         do 820 j = 1, 3
            R(i,j) = 0.0d0
  820    continue
  830 continue
      do 850 i=1,3
         R(i,i) = 1.0d0
  850 continue
      go to 4550
c
c     optimal rotation is taken to be F matrix (rotation for symmetry)
c
  900 continue
      ICOUNT3 = ICOUNT3+1
      do 930 i = 1, 3
         do 920 j = 1, 3
            R(i,j) = F(i,j)
            A(i,j) = M(i,j)
  920    continue
  930 continue
c     write(*,*)'Matrix R:'
c     do 960 i = 1, 3
c        write(*,*)(R(i,j),j=1,3)
c 960 continue
c****************************
c***********************************
      go to 4550
c*****************************
c
 1000 continue
      ICOUNT4 = ICOUNT4+1
c     write(*,*)'icode=',icode
c     write(*,*)'Matrix A is symmetric but matrix S does not appear'
c     write(*,*)'to be positive semidefinite'
c     write(*,*)'Thus matrix A is not of maximal trace over rotations'
c     write(*,*)'Program proceeds to obtain rotation matrix R so that'
c     write(*,*)'the product RA is of maximal trace over rotations'
c     write(*,*)'and this is done without using the SVD concept'
c
c     Matrix A is not of maximal trace over rotations
c     but it is symmetric
c     Now program proceeds to obtain rotation matrix R so that
c     the product RA is of maximal trace over rotations and
c     this is done without using the SVD concept
c
      call eigval(A, sig1, sig2, sig3, epp)
      call eigvec(A, sig3, r3, epp, epn)
c
c     compute optimal rotation
c
      do 1300 i = 1, 3
         do 1200 j = 1, 3
            S(i,j) = 2.0d0*r3(i)*r3(j)
 1200    continue
 1300 continue
      do 1500 i=1,3
         S(i,i) = S(i,i) - 1.0d0
 1500 continue
      if(iflag.ne.0) then
         do 1530 i = 1, 3
            do 1520 j = 1, 3
               A(i,j) = M(i,j)
               R(i,j) = 0.0d0
               do 1510 k = 1, 3
                  R(i,j) = R(i,j) + S(i,k)*F(k,j)
 1510          continue
 1520       continue
 1530    continue
      else
         do 1550 i = 1, 3
            do 1540 j = 1, 3
               R(i,j) = S(i,j)
 1540       continue
 1550   continue
      endif
c     write(*,*)'Matrix R:'
c     do 1600 i = 1, 3
c        write(*,*)(R(i,j),j=1,3)
c1600 continue
c
c***************
c**************************
      go to 4550
c***********************************
c
 2000 continue
      if(svdonly.eq.1) go to 3000
c     write(*,*)'Matrix A does not appear to be symmetric'
c     write(*,*)'Program proceeds to attempt to obtain rotation'
c     write(*,*)'matrix U so that UA is symmetric using the Cayley'
c     write(*,*)'transform in conjuction with Newton's method'
c     write(*,*)'If successful then program proceeds to obtain matrix'
c     write(*,*)'R so that RUA is of maximal trace over rotations'
c     write(*,*)'without using the SVD concept'
c     write(*,*)'If not successful then program proceeds to obtain'
c     write(*,*)'rotation matrix R so that RA is of maximal trace'
c     write(*,*)'over rotations using the SVD concept on (A^T)A which'
c     write(*,*)'is indeed a symmetric matrix'
c
c     Matrix A does not appear to be symmetric
c     Program proceeds to attempt to obtain rotation matrix U so that
c     UA is symmetric using the Cayley transform in conjuction with
c     Newton's method
c     If successful then program proceeds to obtain matrix R so that
c     RUA is of maximal trace over rotations without using the SVD
c     concept
c     If not successful then program proceeds to obtain rotation matrix
c     R so that RA is of maximal trace over rotations using the SVD
c     concept on (A^T)A which is indeed a symmetric matrix
c
      rr = 0.0d0
      ss = 0.0d0
      tt = 0.0d0
 2050 continue
c     WRITE(*,*)'RR SS TT=',RR,SS,TT
      iter = iter + 1
      rs = rr*ss
      rt = rr*tt
      st = ss*tt
      rr2 = rr**2
      ss2 = ss**2
      tt2 = tt**2
      delta2 = (1.0d0 + rr2 + ss2 + tt2)/2.0d0
      F(1,1) = delta2 - rr2 - ss2
      F(1,2) = -rr + st
      F(1,3) = ss + rt
      F(2,1) = rr + st
      F(2,2) = delta2 - rr2 - tt2
      F(2,3) = -tt + rs
      F(3,1) = -ss + rt
      F(3,2) = tt + rs
      F(3,3) = delta2 - ss2 - tt2
      g(1) = F(1,1)*A(1,2) + F(1,2)*A(2,2) + F(1,3)*A(3,2)
     *     - A(1,1)*F(2,1) - A(2,1)*F(2,2) - A(3,1)*F(2,3)
      g(2) = F(3,1)*A(1,1) + F(3,2)*A(2,1) + F(3,3)*A(3,1)
     *     - A(1,3)*F(1,1) - A(2,3)*F(1,2) - A(3,3)*F(1,3)
      g(3) = F(2,1)*A(1,3) + F(2,2)*A(2,3) + F(2,3)*A(3,3)
     *     - A(1,2)*F(3,1) - A(2,2)*F(3,2) - A(3,2)*F(3,3)
c     WRITE(*,*)'g123 =',(g(i),i=1,3)
      if(g(1).gt.epn.and.g(1).lt.epp .and.
     *   g(2).gt.epn.and.g(2).lt.epp .and.
     *   g(3).gt.epn.and.g(3).lt.epp) go to 2300
      if(iter.ge.itex) go to 3000
      Fr(1,1) = -rr
      Fr(1,2) = -1
      Fr(1,3) = tt
      Fr(2,1) = 1
      Fr(2,2) = -rr
      Fr(2,3) = ss
      Fr(3,1) = tt
      Fr(3,2) = ss
      Fr(3,3) = rr
      Fs(1,1) = -ss
      Fs(1,2) = tt
      Fs(1,3) = 1
      Fs(2,1) = tt
      Fs(2,2) = ss
      Fs(2,3) = rr
      Fs(3,1) = -1
      Fs(3,2) = rr
      Fs(3,3) = -ss
      Ft(1,1) = tt
      Ft(1,2) = ss
      Ft(1,3) = rr
      Ft(2,1) = ss
      Ft(2,2) = -tt
      Ft(2,3) = -1
      Ft(3,1) = rr
      Ft(3,2) = 1
      Ft(3,3) = -tt
c
c     compute Jacobian of g
c
      JJ(1,1) = Fr(1,1)*A(1,2) + Fr(1,2)*A(2,2) + Fr(1,3)*A(3,2)
     *       - A(1,1)*Fr(2,1) - A(2,1)*Fr(2,2) - A(3,1)*Fr(2,3)
      JJ(2,1) = Fr(3,1)*A(1,1) + Fr(3,2)*A(2,1) + Fr(3,3)*A(3,1)
     *       - A(1,3)*Fr(1,1) - A(2,3)*Fr(1,2) - A(3,3)*Fr(1,3)
      JJ(3,1) = Fr(2,1)*A(1,3) + Fr(2,2)*A(2,3) + Fr(2,3)*A(3,3)
     *       - A(1,2)*Fr(3,1) - A(2,2)*Fr(3,2) - A(3,2)*Fr(3,3)
      JJ(1,2) = Fs(1,1)*A(1,2) + Fs(1,2)*A(2,2) + Fs(1,3)*A(3,2)
     *       - A(1,1)*Fs(2,1) - A(2,1)*Fs(2,2) - A(3,1)*Fs(2,3)
      JJ(2,2) = Fs(3,1)*A(1,1) + Fs(3,2)*A(2,1) + Fs(3,3)*A(3,1)
     *       - A(1,3)*Fs(1,1) - A(2,3)*Fs(1,2) - A(3,3)*Fs(1,3)
      JJ(3,2) = Fs(2,1)*A(1,3) + Fs(2,2)*A(2,3) + Fs(2,3)*A(3,3)
     *       - A(1,2)*Fs(3,1) - A(2,2)*Fs(3,2) - A(3,2)*Fs(3,3)
      JJ(1,3) = Ft(1,1)*A(1,2) + Ft(1,2)*A(2,2) + Ft(1,3)*A(3,2)
     *       - A(1,1)*Ft(2,1) - A(2,1)*Ft(2,2) - A(3,1)*Ft(2,3)
      JJ(2,3) = Ft(3,1)*A(1,1) + Ft(3,2)*A(2,1) + Ft(3,3)*A(3,1)
     *       - A(1,3)*Ft(1,1) - A(2,3)*Ft(1,2) - A(3,3)*Ft(1,3)
      JJ(3,3) = Ft(2,1)*A(1,3) + Ft(2,2)*A(2,3) + Ft(2,3)*A(3,3)
     *       - A(1,2)*Ft(3,1) - A(2,2)*Ft(3,2) - A(3,2)*Ft(3,3)
c     write(*,*)'Matrix JJ:'
c     do 2100 i = 1, 3
c        write(*,*)(JJ(i,j),j=1,3)
c2100 continue
      d3 = JJ(1,1)*JJ(2,2)*JJ(3,3) + JJ(2,1)*JJ(3,2)*JJ(1,3) +
     *     JJ(1,2)*JJ(2,3)*JJ(3,1) - JJ(1,3)*JJ(2,2)*JJ(3,1) -
     *     JJ(1,2)*JJ(2,1)*JJ(3,3) - JJ(2,3)*JJ(3,2)*JJ(1,1)
c     WRITE(*,*)'D3=',D3
      if(d3.gt.epn .and. d3.lt.epp) go to 3000
c
c     compute inverse of Jacobian
c
      JI(1,1) = JJ(2,2)*JJ(3,3)-JJ(3,2)*JJ(2,3)
      JI(1,2) = JJ(3,2)*JJ(1,3)-JJ(1,2)*JJ(3,3)
      JI(1,3) = JJ(1,2)*JJ(2,3)-JJ(2,2)*JJ(1,3)
      JI(2,1) = JJ(2,3)*JJ(3,1)-JJ(3,3)*JJ(2,1)
      JI(2,2) = JJ(3,3)*JJ(1,1)-JJ(1,3)*JJ(3,1)
      JI(2,3) = JJ(1,3)*JJ(2,1)-JJ(2,3)*JJ(1,1)
      JI(3,1) = JJ(2,1)*JJ(3,2)-JJ(3,1)*JJ(2,2)
      JI(3,2) = JJ(3,1)*JJ(1,2)-JJ(1,1)*JJ(3,2)
      JI(3,3) = JJ(1,1)*JJ(2,2)-JJ(2,1)*JJ(1,2)
c     write(*,*)'Matrix JI:'
c     do 2200 i = 1, 3
c        write(*,*)(JI(i,j),j=1,3)
c2200 continue
      h(1) = (JI(1,1)*g(1) + JI(1,2)*g(2) + JI(1,3)*g(3))/d3
      h(2) = (JI(2,1)*g(1) + JI(2,2)*g(2) + JI(2,3)*g(3))/d3
      h(3) = (JI(3,1)*g(1) + JI(3,2)*g(2) + JI(3,3)*g(3))/d3
      rr = rr - h(1)
      ss = ss - h(2)
      tt = tt - h(3)
      go to 2050
 2300 continue
      ICOUNT1 = ICOUNT1+1
      iflag = 1
      do 2500 i = 1, 3
         do 2400 j = 1, 3
            M(i,j) = A(i,j)
            F(i,j) = F(i,j)/delta2
 2400    continue
 2500 continue
c     d3 = F(1,1)*F(2,2)*F(3,3) + F(2,1)*F(3,2)*F(1,3) +
c    *     F(1,2)*F(2,3)*F(3,1) - F(1,3)*F(2,2)*F(3,1) -
c    *     F(1,2)*F(2,1)*F(3,3) - F(2,3)*F(3,2)*F(1,1)
c     WRITE(*,*)'**** D3=',D3
      do 2800 i = 1, 3
         do 2700 j = 1, 3
            A(i,j) = 0.0d0
            do 2600 k = 1, 3
               A(i,j) = A(i,j) + F(i,k)*M(k,j)
 2600       continue
 2700    continue
 2800 continue
c     write(*,*)'2900 Matrix A:'
c     do 2900 i = 1, 3
c        write(*,*)(A(i,j),j=1,3)
c2900 continue
      go to 120
c
c
c     Cayley transform Newton method not successful
c     Program proceeds to obtain rotation matrix R so that RA is
c     of maximal trace over rotations using the SVD concept on
c     (A^T)A which is indeed a symmetric matrix
c
c
c     compute (A^T)A
c
 3000 continue
      ICOUNT2 = ICOUNT2+1
      do 3300 i = 1, 3
         do 3200 j = 1, 3
            ATA(i,j) = 0.0d0
            do 3100 k = 1, 3
               ATA(i,j) = ATA(i,j) + A(k,i)*A(k,j)
 3100       continue
 3200    continue
 3300 continue
c     write(*,*)'Matrix ATA:'
c     do 3400 i = 1, 3
c        write(*,*)(ATA(i,j),j=1,3)
c3400 continue
c
      call eigval(ATA, sig1, sig2, sig3, epp)
      if(sig1.lt.epp) sig1 = 0.0d0
      if(sig2.lt.epp) sig2 = 0.0d0
      if(sig3.lt.epp) sig3 = 0.0d0
      sig(1) = sig3
      sig(2) = sig2
      sig(3) = sig1
c
c     compute V matrix for SVD
c
      call eigves(ATA, sig, w, epp, epn)
c
c     compute U matrix for SVD
c
      call umxves(A, sig, w, u, epp)
c     WRITE(*,*)'SINGULAR VALUES =',(SIG(i),i=1,3)
c
c     test SVD result
c
      do 3600 i = 1, 3
         do 3500 j = 1, 3
            S(i,j) = sig(i)*u(i,j)
 3500    continue
 3600 continue
      do 3900 i = 1, 3
         do 3800 j = 1, 3
            ATA(i,j) = 0.0d0
            do 3700 k = 1, 3
               ATA(i,j) = ATA(i,j) + S(k,i)*w(k,j)
 3700       continue
 3800    continue
 3900 continue
c     write(*,*)'SVD Matrix A:'
c     do 4000 i = 1, 3
c        write(*,*)(ATA(i,j),j=1,3)
c4000 continue
c     write(*,*)'Matrix A:'
c     do 4100 i = 1, 3
c        write(*,*)(A(i,j),j=1,3)
c4100 continue
c
c     compute optimal rotation
c
      d2 = u(1,1)*u(2,2)*u(3,3) + u(2,1)*u(3,2)*u(1,3) +
     *     u(1,2)*u(2,3)*u(3,1) - u(1,3)*u(2,2)*u(3,1) -
     *     u(1,2)*u(2,1)*u(3,3) - u(2,3)*u(3,2)*u(1,1)
      d3 = w(1,1)*w(2,2)*w(3,3) + w(2,1)*w(3,2)*w(1,3) +
     *     w(1,2)*w(2,3)*w(3,1) - w(1,3)*w(2,2)*w(3,1) -
     *     w(1,2)*w(2,1)*w(3,3) - w(2,3)*w(3,2)*w(1,1)
c     WRITE(*,*)'D23=',D2,D3
      if(d2*d3.lt.0.0d0) then
         w(3,1) = -w(3,1)
         w(3,2) = -w(3,2)
         w(3,3) = -w(3,3)
      endif
      do 4400 i = 1, 3
         do 4300 j = 1, 3
            R(i,j) = 0.0d0
            do 4200 k = 1, 3
               R(i,j) = R(i,j) + w(k,i)*u(k,j)
 4200       continue
 4300    continue
 4400 continue
c     write(*,*)'Matrix R:'
c     do 4500 i = 1, 3
c        write(*,*)(R(i,j),j=1,3)
c4500 continue
c
 4550 continue
      return
      end
*EIGVAL
c
c     subroutine eigval to
c
c     compute eigenvalues sig1, sig2, sig3 of symmetric matrix A,
c     sig1 <= sig2 <= sig3 
c
      subroutine eigval(A, sig1, sig2, sig3, epp)
c
      double precision A(3,3), B(3,3)
      double precision sig1, sig2, sig3, epp
      double precision qt, pt, detB, twoPI
c
      qt = (A(1,1)+A(2,2)+A(3,3))/3.0d0
c     write(*,*)'qt =',qt
c
      pt = dsqrt(((A(1,1)-qt)**2 + (A(2,2)-qt)**2 + (A(3,3)-qt)**2 +
     * 2.0d0*A(1,2)**2 + 2.0d0*A(1,3)**2 + 2.0d0*A(2,3)**2)/6.0d0)
c     write(*,*)'pt =',pt
c
c     A is a diagonal matrix with the same number in each entry of the
c     diagonal which is the only eigenvalue of A (of multiplicity 3)
c     Since A is not of maximal trace over rotation matrices this
c     eigenvalue (the only one) must be negative
c
      if(pt.ge.epp) go to 100
      sig1 = qt
      sig2 = qt
      sig3 = qt
      go to 2000
c
  100 continue
c
c     A is not a trivial matrix
c     set up matrix B = (A - qt*I)/pt and find eigenvalues of A
c
      do 300 j = 1, 3
         do 200 i = 1, 3
            B(i,j) = A(i,j)
  200    continue
  300 continue
      do 500 i=1,3
         B(i,i) = B(i,i) - qt
  500 continue
      do 700 j = 1, 3
         do 600 i = 1, 3
            B(i,j) = B(i,j)/pt
  600    continue
  700 continue
c     write(*,*)'Matrix B:'
c     do 800 i = 1, 3
c        write(*,*)(B(i,j),j=1,3)
c 800 continue
c
      detB = B(1,1)*B(2,2)*B(3,3) + B(2,1)*B(3,2)*B(1,3) +
     *     B(1,2)*B(2,3)*B(3,1) - B(1,3)*B(2,2)*B(3,1) -
     *     B(1,2)*B(2,1)*B(3,3) - B(2,3)*B(3,2)*B(1,1)
c     WRITE(*,*)'detB=',detB
c
      twoPI = 2.0d0*dacos(-1.0d0)
      if(detB.ge.2.0d0) then
         sig3 = dacos(1.0d0)
      elseif(detB.le.-2.0d0) then
         sig3 = dacos(-1.0d0)
      else
         sig3 = dacos(detB/2.0d0)
      endif
      sig2 = 2*pt*cos((twoPI - sig3)/3.0d0) + qt
      sig1 = 2*pt*cos((twoPI + sig3)/3.0d0) + qt
      sig3 = 2*pt*cos(sig3/3.0d0) + qt
c     WRITE(*,*)'SIG123=',SIG1,SIG2,SIG3
c
 2000 continue
c
      return
      end
*EIGVEC
c
c     subroutine eigvec to
c
c     compute unit eigenvector of A corresponding to
c     eigenvalue (sig) of A
c
      subroutine eigvec(A, sig, r, epp, epn)
c
      double precision A(3,3), C(3,3), r(3), v(3,3)
      double precision dlen(3), dmax, sig, epp, epn
      integer icol(3), ivol(3), imax, i, j, ipos
c
c     set up matrix C = A - sig*I
c
      do 300 j = 1, 3
         do 200 i = 1, 3
            C(i,j) = A(i,j)
  200    continue
  300 continue
      do 500 i=1,3
         C(i,i) = C(i,i) - sig
  500 continue
c     write(*,*)'Matrix C:'
c     do 550 i = 1, 3
c        write(*,*)(C(i,j),j=1,3)
c 550 continue
c
c     test for nonzero columns
c
      ipos = 0
      do 700 j = 1, 3
         ipos = ipos + 1
         icol(j) = 1
         do 600 i = 1, 3
            if(C(i,j).le.epn .or. C(i,j).ge.epp) go to 700
  600    continue
         icol(j) = 0
         ipos = ipos - 1
  700 continue
c
c     WRITE(*,*)'ICOL123=',(ICOL(J),J=1,3)
      if(ipos.eq.1) go to 1400
      if(ipos.ne.0) go to 800
c
c     all columns are zero
c
  750 continue
      r(1) = 0.0d0
      r(2) = 0.0d0
      r(3) = 1.0d0
      go to 9000
c
c     test cross products of columns of A
c
  800 continue
      if(ipos.eq.2) go to 1300
c
c     all columns are nonzero
c
      v(1,1) = C(2,1)*C(3,2)-C(3,1)*C(2,2)
      v(1,2) = C(3,1)*C(1,2)-C(1,1)*C(3,2)
      v(1,3) = C(1,1)*C(2,2)-C(2,1)*C(1,2)
      v(2,1) = C(2,2)*C(3,3)-C(3,2)*C(2,3)
      v(2,2) = C(3,2)*C(1,3)-C(1,2)*C(3,3)
      v(2,3) = C(1,2)*C(2,3)-C(2,2)*C(1,3)
      v(3,1) = C(2,3)*C(3,1)-C(3,3)*C(2,1)
      v(3,2) = C(3,3)*C(1,1)-C(1,3)*C(3,1)
      v(3,3) = C(1,3)*C(2,1)-C(2,3)*C(1,1)
c
      do 900 i = 1, 3
         ivol(i) = 1
         if(v(i,1).gt.epn .and. v(i,1).lt.epp .and. v(i,2).gt.epn .and.
     *      v(i,2).lt.epp .and. v(i,3).gt.epn .and. v(i,3).lt.epp)
     *      ivol(i)= 0
  900 continue
c     WRITE(*,*)'IVOL123=',(IVOL(J),J=1,3)
      
      if(ivol(1).eq.0 .and. ivol(2).eq.0 .and. ivol(3).eq.0) go to 1400
c
c      first test passed: there are nonzero cross products of columns
c      column space appears 2-dimensional
c
      do 1000 i = 1, 3
         dlen(i) = 0.0d0
         if(ivol(i).ne.0)
     *      dlen(i) = dsqrt(v(i,1)**2 + v(i,2)**2 + v(i,3)**2)
 1000 continue
      do 1100 i = 1, 3
         if(dlen(i).ge.epp) go to 1200
 1100 continue
      go to 1400
 1200 continue
c
c     second test passed: there are nonzero cross products of columns
c     column space is 2-dimensional
c     cross product with largest length is taken as eigenvector
c
      dmax = -10.0d0
      imax = 0
      do 1250 i = 1, 3
         if(dlen(i).le.dmax) go to 1250
         dmax = dlen(i)
         imax = i
 1250 continue
      r(1) = v(imax,1)/dmax
      r(2) = v(imax,2)/dmax
      r(3) = v(imax,3)/dmax
c     write(*,*)'1200 imax=',imax,' dmax=',dmax,' r123=',(r(i),i=1,3)
      go to 9000
c
c     one column is zero, two columns are nonzero
c
 1300 continue
      if(icol(3).eq.0) then
         r(1) = C(2,1)*C(3,2)-C(3,1)*C(2,2)
         r(2) = C(3,1)*C(1,2)-C(1,1)*C(3,2)
         r(3) = C(1,1)*C(2,2)-C(2,1)*C(1,2)
      elseif(icol(1).eq.0) then
         r(1) = C(2,2)*C(3,3)-C(3,2)*C(2,3)
         r(2) = C(3,2)*C(1,3)-C(1,2)*C(3,3)
         r(3) = C(1,2)*C(2,3)-C(2,2)*C(1,3)
      else
         r(1) = C(2,3)*C(3,1)-C(3,3)*C(2,1)
         r(2) = C(3,3)*C(1,1)-C(1,3)*C(3,1)
         r(3) = C(1,3)*C(2,1)-C(2,3)*C(1,1)
      endif
      if(r(1).gt.epn .and. r(1).lt.epp .and. r(2).gt.epn .and.
     *   r(2).lt.epp .and. r(3).gt.epn .and. r(3).lt.epp) go to 1400
      dmax = dsqrt(r(1)**2 + r(2)**2 + r(3)**2)
      if(dmax.lt.epp) go to 1400
      r(1) = r(1)/dmax
      r(2) = r(2)/dmax
      r(3) = r(3)/dmax
c     write(*,*)'1300 dmax=',dmax,' r123=',(r(i),i=1,3)
      go to 9000
c
c     column space is 1-dimensional
c
 1400 continue
      dmax = -10.0d0
      imax = 0
      do 1500 i = 1, 3
         dlen(i) = 0.0d0
         if(icol(i).ne.0)
     *      dlen(i) = dsqrt(C(1,i)**2 + C(2,i)**2 + C(3,i)**2)
         if(dlen(i).le.dmax) go to 1500
         dmax = dlen(i)
         imax = i
 1500 continue
      if(dmax.lt.epp) go to 750
      v(1,1) = C(1,imax)/dmax
      v(1,2) = C(2,imax)/dmax
      v(1,3) = C(3,imax)/dmax
c
      v(2,1) = 1.0d0
      v(2,2) = 1.0d0
      v(2,3) = 1.0d0
c
      dmax = dabs(v(1,1))
      imax = 1
      if(dabs(v(1,2)).gt.dmax) then
         dmax = dabs(v(1,2))
         imax = 2
      endif
      if(dabs(v(1,3)).gt.dmax) then
         dmax = dabs(v(1,3))
         imax = 3
      endif
      v(2,imax) = 0.0d0
c
      r(1) = v(1,2)*v(2,3) - v(1,3)*v(2,2)
      r(2) = v(1,3)*v(2,1) - v(1,1)*v(2,3)
      r(3) = v(1,1)*v(2,2) - v(1,2)*v(2,1)
      dmax = dsqrt(r(1)**2 + r(2)**2 + r(3)**2)
      if(dmax.lt.epp) go to 750
      r(1) = r(1)/dmax
      r(2) = r(2)/dmax
      r(3) = r(3)/dmax
c     write(*,*)'1500 dmax=',dmax,' r123=',(r(i),i=1,3)
c
 9000 continue
c
c     test eigenvector
c
      v(1,1) = C(1,1)*r(1) + C(1,2)*r(2) + C(1,3)*r(3)
      v(1,2) = C(2,1)*r(1) + C(2,2)*r(2) + C(2,3)*r(3)
      v(1,3) = C(3,1)*r(1) + C(3,2)*r(2) + C(3,3)*r(3)
c     WRITE(*,*)'V123=',(V(1,I),I=1,3)
c
      return
      end
*EIGVES
c
c     subroutine eigves to
c
c     compute orthonormal set of eigenvectors of A that span R^3
c     used mainly to compute set of vectors for V matrix in SVD
c
      subroutine eigves(A, sig, w, epp, epn)
c
      double precision A(3,3), C(3,3), w(3,3), v(3,3), sig(3)
      double precision dlen(3), dmax, epp, epn
      integer icol(3), ivol(3), imax, i, j, k, ipos
c
c     set up matrix C = A - sig(k)*I
c
      k = 1
  100 continue
      do 300 j = 1, 3
         do 200 i = 1, 3
            C(i,j) = A(i,j)
  200    continue
  300 continue
      do 500 i=1,3
         C(i,i) = C(i,i) - sig(k)
  500 continue
c     write(*,*)'Matrix C:'
c     do 550 i = 1, 3
c        write(*,*)(C(i,j),j=1,3)
c 550 continue
c
c     test for nonzero columns
c
      ipos = 0
      do 700 j = 1, 3
         ipos = ipos + 1
         icol(j) = 1
         do 600 i = 1, 3
            if(C(i,j).le.epn .or. C(i,j).ge.epp) go to 700
  600    continue
         icol(j) = 0
         ipos = ipos - 1
  700 continue
c
c     WRITE(*,*)'ICOL123=',(ICOL(J),J=1,3)
      if(ipos.eq.1) go to 1400
      if(ipos.ne.0) go to 800
c
c     all columns are zero
c
  750 continue
      w(1,1) = 1.0d0
      w(1,2) = 0.0d0
      w(1,3) = 0.0d0
      w(2,1) = 0.0d0
      w(2,2) = 1.0d0
      w(2,3) = 0.0d0
      w(3,1) = 0.0d0
      w(3,2) = 0.0d0
      w(3,3) = 1.0d0
c     write(*,*)'750-1 w123=',(w(1,i),i=1,3)
c     write(*,*)'750-2 w123=',(w(2,i),i=1,3)
c     write(*,*)'750-3 w123=',(w(3,i),i=1,3)
c
c     test eigenvectors
c
c     WRITE(*,*)'750 TESTING EIGENVECTORS'
      do 770 k = 1, 3
         v(k,1) = C(1,1)*w(k,1) + C(1,2)*w(k,2) + C(1,3)*w(k,3)
         v(k,2) = C(2,1)*w(k,1) + C(2,2)*w(k,2) + C(2,3)*w(k,3)
         v(k,3) = C(3,1)*w(k,1) + C(3,2)*w(k,2) + C(3,3)*w(k,3)
c        WRITE(*,*)'k=',k,' V123=',(V(k,I),I=1,3)
  770 continue
      go to 9000
c
c     test cross products of columns of A
c
  800 continue
      if(ipos.eq.2) go to 1300
c
c     all columns are nonzero
c
      v(1,1) = C(2,1)*C(3,2)-C(3,1)*C(2,2)
      v(1,2) = C(3,1)*C(1,2)-C(1,1)*C(3,2)
      v(1,3) = C(1,1)*C(2,2)-C(2,1)*C(1,2)
      v(2,1) = C(2,2)*C(3,3)-C(3,2)*C(2,3)
      v(2,2) = C(3,2)*C(1,3)-C(1,2)*C(3,3)
      v(2,3) = C(1,2)*C(2,3)-C(2,2)*C(1,3)
      v(3,1) = C(2,3)*C(3,1)-C(3,3)*C(2,1)
      v(3,2) = C(3,3)*C(1,1)-C(1,3)*C(3,1)
      v(3,3) = C(1,3)*C(2,1)-C(2,3)*C(1,1)
c
      do 900 i = 1, 3
         ivol(i) = 1
         if(v(i,1).gt.epn .and. v(i,1).lt.epp .and. v(i,2).gt.epn .and.
     *      v(i,2).lt.epp .and. v(i,3).gt.epn .and. v(i,3).lt.epp)
     *      ivol(i)= 0
  900 continue
c     WRITE(*,*)'IVOL123=',(IVOL(J),J=1,3)
      if(ivol(1).eq.0 .and. ivol(2).eq.0 .and. ivol(3).eq.0) go to 1400
c
c      first test passed: there are nonzero cross products of columns
c      column space appears 2-dimensional
c
      do 1000 i = 1, 3
         dlen(i) = 0.0d0
         if(ivol(i).ne.0)
     *      dlen(i) = dsqrt(v(i,1)**2 + v(i,2)**2 + v(i,3)**2)
 1000 continue
      do 1100 i = 1, 3
         if(dlen(i).ge.epp) go to 1200
 1100 continue
      go to 1400
 1200 continue
c
c     second test passed: there are nonzero cross products of columns
c     column space is 2-dimensional
c     cross product with largest length is taken as eigenvector
c
      dmax = -10.0d0
      imax = 0
      do 1250 i = 1, 3
         if(dlen(i).le.dmax) go to 1250
         dmax = dlen(i)
         imax = i
 1250 continue
      w(k,1) = v(imax,1)/dmax
      w(k,2) = v(imax,2)/dmax
      w(k,3) = v(imax,3)/dmax
c     write(*,*)'1200 dmax=',dmax,' k=',k,' w123=',(w(k,i),i=1,3)
      go to 1320
c
c     one column is zero, two columns are nonzero
c
 1300 continue
      if(icol(3).eq.0) then
         w(k,1) = C(2,1)*C(3,2)-C(3,1)*C(2,2)
         w(k,2) = C(3,1)*C(1,2)-C(1,1)*C(3,2)
         w(k,3) = C(1,1)*C(2,2)-C(2,1)*C(1,2)
      elseif(icol(1).eq.0) then
         w(k,1) = C(2,2)*C(3,3)-C(3,2)*C(2,3)
         w(k,2) = C(3,2)*C(1,3)-C(1,2)*C(3,3)
         w(k,3) = C(1,2)*C(2,3)-C(2,2)*C(1,3)
      else
         w(k,1) = C(2,3)*C(3,1)-C(3,3)*C(2,1)
         w(k,2) = C(3,3)*C(1,1)-C(1,3)*C(3,1)
         w(k,3) = C(1,3)*C(2,1)-C(2,3)*C(1,1)
      endif
      if(w(k,1).gt.epn .and. w(k,1).lt.epp .and. w(k,2).gt.epn .and.
     *   w(k,2).lt.epp .and. w(k,3).gt.epn .and. w(k,3).lt.epp)
     *   go to 1400
      dmax = dsqrt(w(k,1)**2 + w(k,2)**2 + w(k,3)**2)
      if(dmax.lt.epp) go to 1400
      w(k,1) = w(k,1)/dmax
      w(k,2) = w(k,2)/dmax
      w(k,3) = w(k,3)/dmax
c     write(*,*)'1300 dmax=',dmax,' k=',k,' w123=',(w(k,i),i=1,3)
c
 1320 continue
      if(k.eq.3) go to 1350
      k = 3 
      go to 100
c
c     compute eigenvector for sig(2)
c
 1350 continue
      k = 2
      w(2,1) = w(1,2)*w(3,3) - w(1,3)*w(3,2)
      w(2,2) = w(1,3)*w(3,1) - w(1,1)*w(3,3)
      w(2,3) = w(1,1)*w(3,2) - w(1,2)*w(3,1)
      dmax = dsqrt(w(2,1)**2 + w(2,2)**2 + w(2,3)**2)
      if(dmax.lt.epp) go to 1400
      w(2,1) = w(2,1)/dmax
      w(2,2) = w(2,2)/dmax
      w(2,3) = w(2,3)/dmax
c     write(*,*)'1350 dmax=',dmax,' k=',k,' w123=',(w(k,i),i=1,3)
c
c     test eigenvectors
c
c     WRITE(*,*)'1350 TESTING EIGENVECTORS'
      do 1370 k = 1, 3
         do 1360 i=1,3
            C(i,i) = A(i,i) - sig(k)
 1360    continue
         v(k,1) = C(1,1)*w(k,1) + C(1,2)*w(k,2) + C(1,3)*w(k,3)
         v(k,2) = C(2,1)*w(k,1) + C(2,2)*w(k,2) + C(2,3)*w(k,3)
         v(k,3) = C(3,1)*w(k,1) + C(3,2)*w(k,2) + C(3,3)*w(k,3)
c        WRITE(*,*)'k=',k,' V123=',(V(k,I),I=1,3)
 1370 continue
      go to 9000
c
c     column space is 1-dimensional
c     sig(2) is of multiplicity 2
c
 1400 continue
      dmax = -10.0d0
      imax = 0
      do 1500 i = 1, 3
         dlen(i) = 0.0d0
         if(icol(i).ne.0)
     *      dlen(i) = dsqrt(C(1,i)**2 + C(2,i)**2 + C(3,i)**2)
         if(dlen(i).le.dmax) go to 1500
         dmax = dlen(i)
         imax = i
 1500 continue
      if(dmax.lt.epp) go to 750
      v(1,1) = C(1,imax)/dmax
      v(1,2) = C(2,imax)/dmax
      v(1,3) = C(3,imax)/dmax
c     write(*,*)'1500-0 dmax=',dmax,' v123=',(v(1,i),i=1,3)
c
      v(2,1) = 1.0d0
      v(2,2) = 1.0d0
      v(2,3) = 1.0d0
c
      dmax = dabs(v(1,1))
      imax = 1
      if(dabs(v(1,2)).gt.dmax) then
         dmax = dabs(v(1,2))
         imax = 2
      endif
      if(dabs(v(1,3)).gt.dmax) then
         dmax = dabs(v(1,3))
         imax = 3
      endif
      v(2,imax) = 0.0d0
c
      w(k,1) = v(1,2)*v(2,3) - v(1,3)*v(2,2)
      w(k,2) = v(1,3)*v(2,1) - v(1,1)*v(2,3)
      w(k,3) = v(1,1)*v(2,2) - v(1,2)*v(2,1)
      dmax = dsqrt(w(k,1)**2 + w(k,2)**2 + w(k,3)**2)
      if(dmax.lt.epp) go to 750
      w(k,1) = w(k,1)/dmax
      w(k,2) = w(k,2)/dmax
      w(k,3) = w(k,3)/dmax
c     write(*,*)'1500-1 dmax=',dmax,' k=',k,' w123=',(w(k,i),i=1,3)
c
c     compute another eigenvector for sig(2)
c
      w(2,1) = w(k,2)*v(1,3) - w(k,3)*v(1,2)
      w(2,2) = w(k,3)*v(1,1) - w(k,1)*v(1,3)
      w(2,3) = w(k,1)*v(1,2) - w(k,2)*v(1,1)
      dmax = dsqrt(w(2,1)**2 + w(2,2)**2 + w(2,3)**2)
      if(dmax.lt.epp) go to 750
      w(2,1) = w(2,1)/dmax
      w(2,2) = w(2,2)/dmax
      w(2,3) = w(2,3)/dmax
      j = k
      k = 2
c     write(*,*)'1500-2 dmax=',dmax,' k=',k,' w123=',(w(k,i),i=1,3)
c
c     compute eigenvector for eigenvalue not equal to sig(2)
c
      if(j.eq.1) then
         k = 3
      else
         k = 1
      endif
c     w(k,1) = w(2,2)*w(j,3) - w(2,3)*w(j,2)
c     w(k,2) = w(2,3)*w(j,1) - w(2,1)*w(j,3)
c     w(k,3) = w(2,1)*w(j,2) - w(2,2)*w(j,1)
c     dmax = dsqrt(w(k,1)**2 + w(k,2)**2 + w(k,3)**2)
c     if(dmax.lt.epp) go to 750
c     w(k,1) = w(k,1)/dmax
c     w(k,2) = w(k,2)/dmax
c     w(k,3) = w(k,3)/dmax
      w(k,1) = v(1,1)
      w(k,2) = v(1,2)
      w(k,3) = v(1,3)
c     write(*,*)'1500-3 k=',k,' w123=',(w(k,i),i=1,3)
c
c     test eigenvectors
c
c     WRITE(*,*)'1500 TESTING EIGENVECTORS'
      do 1600 k = 1, 3
         do 1560 i=1,3
            C(i,i) = A(i,i) - sig(k)
 1560    continue
         v(k,1) = C(1,1)*w(k,1) + C(1,2)*w(k,2) + C(1,3)*w(k,3)
         v(k,2) = C(2,1)*w(k,1) + C(2,2)*w(k,2) + C(2,3)*w(k,3)
         v(k,3) = C(3,1)*w(k,1) + C(3,2)*w(k,2) + C(3,3)*w(k,3)
c        WRITE(*,*)'k=',k,' V123=',(V(k,I),I=1,3)
 1600 continue
c
 9000 continue
c
      return
      end
*UMXVES
c
c     subroutine umxves to
c
c     compute orthonormal set of vectors for U matrix in SVD
c
      subroutine umxves(A, sig, w, u, epp)
c
      double precision A(3,3), w(3,3), u(3,3), sig(3), v(3)
      double precision epp, dmax
      integer imax
c
      if(sig(1).lt.epp) then
         u(1,1) = 1.0d0
         u(1,2) = 0.0d0
         u(1,3) = 0.0d0
         u(2,1) = 0.0d0
         u(2,2) = 1.0d0
         u(2,3) = 0.0d0
         u(3,1) = 0.0d0
         u(3,2) = 0.0d0
         u(3,3) = 1.0d0
         go to 9000
      else
        sig(1) = dsqrt(sig(1))
        u(1,1) = (A(1,1)*w(1,1) + A(1,2)*w(1,2) + A(1,3)*w(1,3))/sig(1)
        u(1,2) = (A(2,1)*w(1,1) + A(2,2)*w(1,2) + A(2,3)*w(1,3))/sig(1)
        u(1,3) = (A(3,1)*w(1,1) + A(3,2)*w(1,2) + A(3,3)*w(1,3))/sig(1)
      endif
c
      if(sig(2).lt.epp) then
         v(1) = 1.0d0
         v(2) = 1.0d0
         v(3) = 1.0d0
c
         dmax = dabs(u(1,1))
         imax = 1
         if(dabs(u(1,2)).gt.dmax) then
            dmax = dabs(u(1,2))
            imax = 2
         endif
         if(dabs(u(1,3)).gt.dmax) then
            dmax = dabs(u(1,3))
            imax = 3
         endif
         v(imax) = 0.0d0
c
         u(2,1) = u(1,2)*v(3) - u(1,3)*v(2)
         u(2,2) = u(1,3)*v(1) - u(1,1)*v(3)
         u(2,3) = u(1,1)*v(2) - u(1,2)*v(1)
         dmax = dsqrt(u(2,1)**2 + u(2,2)**2 + u(2,3)**2)
         if(dmax.lt.epp) stop 100
         u(2,1) = u(2,1)/dmax
         u(2,2) = u(2,2)/dmax
         u(2,3) = u(2,3)/dmax
c
         u(3,1) = u(1,2)*u(2,3) - u(1,3)*u(2,2)
         u(3,2) = u(1,3)*u(2,1) - u(1,1)*u(2,3)
         u(3,3) = u(1,1)*u(2,2) - u(1,2)*u(2,1)
         dmax = dsqrt(u(3,1)**2 + u(3,2)**2 + u(3,3)**2)
         if(dmax.lt.epp) stop 110
         u(3,1) = u(3,1)/dmax
         u(3,2) = u(3,2)/dmax
         u(3,3) = u(3,3)/dmax
         go to 9000
      else
        sig(2) = dsqrt(sig(2))
        u(2,1) = (A(1,1)*w(2,1) + A(1,2)*w(2,2) + A(1,3)*w(2,3))/sig(2)
        u(2,2) = (A(2,1)*w(2,1) + A(2,2)*w(2,2) + A(2,3)*w(2,3))/sig(2)
        u(2,3) = (A(3,1)*w(2,1) + A(3,2)*w(2,2) + A(3,3)*w(2,3))/sig(2)
      endif
c
      if(sig(3).lt.epp) then
         u(3,1) = u(1,2)*u(2,3) - u(1,3)*u(2,2)
         u(3,2) = u(1,3)*u(2,1) - u(1,1)*u(2,3)
         u(3,3) = u(1,1)*u(2,2) - u(1,2)*u(2,1)
         dmax = dsqrt(u(3,1)**2 + u(3,2)**2 + u(3,3)**2)
         if(dmax.lt.epp) stop 110
         u(3,1) = u(3,1)/dmax
         u(3,2) = u(3,2)/dmax
         u(3,3) = u(3,3)/dmax
         go to 9000
      else
        sig(3) = dsqrt(sig(3))
        u(3,1) = (A(1,1)*w(3,1) + A(1,2)*w(3,2) + A(1,3)*w(3,3))/sig(3)
        u(3,2) = (A(2,1)*w(3,1) + A(2,2)*w(3,2) + A(2,3)*w(3,3))/sig(3)
        u(3,3) = (A(3,1)*w(3,1) + A(3,2)*w(3,2) + A(3,3)*w(3,3))/sig(3)
      endif
c
 9000 continue
c     write(*,*)'u123=',(u(1,imax),imax=1,3)
c     write(*,*)'u123=',(u(2,imax),imax=1,3)
c     write(*,*)'u123=',(u(3,imax),imax=1,3)
c
      return
      end
