clear;
%
%     This is Matlab code for reading a sequence of 3x3 matrices, and for
%     each matrix A in the sequence computing a 3x3 rotation matrix R so
%     that RA is of maximal trace.
%     This can be done using Matlab's version of the SVD method
%     only (no Newton's method) or by executing Fortran program maxtrace.f.
%     The execution of maxtrace.f takes place through a Matlab mex file
%     called TD_MEX_MAXTRACE.F of maxtrace.f. Accordingly, a Matlab
%     variable called IFLAG exists in this program for deciding which to
%     use between Matlab's SVD method and the Matlab mex file of maxtrace.f.
%     If IFLAG is set to 1, then the former is used. Otherwise, if IFLAG is
%     not set to~1, say~0, then the latter is used. We note that if IFLAG is
%     not set to~1 so that the Matlab mex file of maxtrace.f is used, then
%     integer variable SVDONLY described in program maxtrace.f must be taken
%     into account as it is part of maxtrace.f.
%
%     See program maxtrace.f for documentation about maxtrace.f.
%
%     Author: Javier Bernal
%             National Institute of Standards and Technology (NIST)
%
%     Disclaimer:
%
%     This software was developed at the National Institute of Standards
%     and Technology by employees of the Federal Government in the
%     course of their official duties. Pursuant to title 17 Section 105
%     of the United States Code this software is not subject to
%     copyright protection and is in the public domain. This software is
%     experimental. NIST assumes no responsibility whatsoever for its
%     use by other parties, and makes no guarantees, expressed or
%     implied, about its quality, reliability, or any other
%     characteristic. We would appreciate acknowledgement if the
%     software is used.
%
tic
f = fopen('randomtrix');
B = fscanf(f, '%g %g %g', [3 inf]);
fclose(f);
%
B = B';
[m,n] = size(B);
inum = m/n;
%
svdonly = 0;
itex = 60;
itar = 0;
ifail = 0;
%
iflag = 1;
%
for k = 1:inum
    fprintf('\n Matrix number = %5.0f \n', k);
    for i = 1:3
        for j = 1:3
            A(i,j) = B((k-1)*3+i,j);
        end
    end
    if iflag 
       [U,S,V] = svd(A);
       detuv = det(U)*det(V);
%   fprintf('\n determinant of U*V = %5.1f \n', detuv);
%
       if det(U*V) > 0
          S = eye(n);
       else
          S = eye(n);
          S(:,end) = -S(:,end);
       end
       R = V*S*U';
       S = R*A;
       trc = trace(S);
       fprintf('\n Trace = %5.5f \n', trc);
       fprintf('\n ***************************** \n');
%
    else
       [R,iter]=TD_MEX_MAXTRACE(A,svdonly,itex);
       if ~svdonly
          if iter>0
             itar = itar + iter;
          else
             ifail = ifail + 1;
          end
       end
%      R
%      iter
       S = R*A;
       trc = trace(S);
       fprintf('\n Trace = %5.5f \n', trc);
       fprintf('\n ***************************** \n');
    end
end
fprintf('\n Total # of matrices = %5.5f \n', inum);
if iflag
    fprintf('\n Only Matlab SVD was used \n');
else
    if svdonly
       fprintf('\n Only Fortran SVD was used \n');
    else
       fprintf('\n Fortran Cayley-Newton procedure was used \n');
       fprintf('\n with Fortran SVD used when Newton method failed \n');
       fprintf('\n # of times Newton method failed = %7.0f \n', ifail);
       isucc = inum-ifail;
       fprintf('\n # of times Newton method succeeded= %7.0f \n', isucc);
       avrg = itar/inum;
       fprintf('\n Total # of Newton iterations = %9.0f \n', itar);
       fprintf('\n Average # of Newton iterations per matrix = %5.5f \n', avrg);
    end
end
fprintf('\n ***************************** \n');
toc
